//------------------------------------------------------------------------------
//            Graph
//------------------------------------------------------------------------------

unit Graph;

interface

uses Windows, Messages, SysUtils;

//------------------------------------------------------------------------------
//           
//  -.      WinAPI
//
//  - ,     WinAPI,   
// (     ,   -    RePaint)
//     ,    
//
//    (3:33 ),   
//
// ,           ,
//       WinAPI  ,
//    -    . ..     
//  .   ,     
// WinAPI
//
//       ,    
//   ClearGraph
//
// F1 -      
//
//    ,  test.dpr
//
// Best Regards, DragoN;
//
//      : dragon@mail.nnov.ru
//------------------------------------------------------------------------------

Const
  DefaultWidth = 640;
  DefaultHeight = 480;

Type
  TColor = TColorRef;

Const
  clBlack = TColor($000000);
  clMaroon = TColor($000080);
  clGreen = TColor($008000);
  clOlive = TColor($008080);
  clNavy = TColor($800000);
  clPurple = TColor($800080);
  clTeal = TColor($808000);
  clGray = TColor($808080);
  clSilver = TColor($C0C0C0);
  clRed = TColor($0000FF);
  clLime = TColor($00FF00);
  clYellow = TColor($00FFFF);
  clBlue = TColor($FF0000);
  clFuchsia = TColor($FF00FF);
  clAqua = TColor($FFFF00);
  clLtGray = TColor($C0C0C0);
  clDkGray = TColor($808080);
  clWhite = TColor($FFFFFF);
  StandardColorsCount = 16;

  clMoneyGreen = TColor($C0DCC0);
  clSkyBlue = TColor($F0CAA6);
  clCream = TColor($F0FBFF);
  clMedGray = TColor($A4A0A0);
  ExtendedColorsCount = 4;

(*   *)

Function SetGraphResolution(w, h: DWord): Boolean;
Function SetGraphToDesktop: Boolean;
Function SetGraphToScreen: Boolean;

Function SetWindowedMode(Windowed: Boolean = True): Boolean;
Function SetOnTopMode(OnTop: Boolean = True): Boolean;
function SetFullScreenMode(W, H: UInt; C: Byte = 0): Bool; OverLoad;
Function SetFullScreenMode(FullScreen: Boolean = True): Boolean; OverLoad;

Function ShowGraphWindow: Boolean;
Function HideGraphWindow: Boolean;

Function WaitGraph: Boolean;
//Function CloseGraph: Boolean;

(*   *)

Function WorkWidth: DWord;  // ~ GetMaxX
Function WorkHeight: DWord; // ~ GetMaxY

function GetPixel(X, Y: Integer): TColorRef;
function SetPixel(X, Y: Integer; Color: TColorRef): TColorRef; // ~ PutPixel

procedure SetLineWidth(Width: DWord);
procedure SetLineStyle(Style: DWord);
(*
PS_SOLID	Pen is solid.
PS_DASH	Pen is dashed.
PS_DOT	Pen is dotted.
PS_DASHDOT	Pen has alternating dashes and dots.
PS_DASHDOTDOT	Pen has dashes and double dots.
PS_NULL	Pen is invisible.
PS_INSIDEFRAME	Pen is solid. When this pen is used in any graphics device interface (GDI) drawing function that takes a bounding rectangle, the dimensions of the figure are shrunk so that it fits entirely in the bounding rectangle, taking into account the width of the pen. This applies only to geometric pens.
*)
procedure SetLineColor(Color: TColorRef);
function GetLineColor: TColorRef;

procedure Line(x1, y1, x2, y2: integer);
procedure Polyline(var Points; Count: Integer);
procedure PolyBezier(const Points; Count: DWORD);

procedure Polygon(var Points; Count: Integer);

procedure Rectangle(x1, y1, x2, y2: integer); // ~ Bar
procedure RoundRect(x1, y1, x2, y2, rx, ry: integer);

procedure Circle(x, y, r: integer);
procedure Ellipse(x, y: integer; xRadius, yRadius: integer);

procedure Arc(X, Y: Integer; xRadius, yRadius: Integer; StartAngle, EndAngle: Single);
procedure Pie(X, Y: Integer; xRadius, yRadius: Integer; StartAngle, EndAngle: Single);
procedure Chord(X, Y: Integer; xRadius, yRadius: Integer; StartAngle, EndAngle: Single);
(* StartAngle, EndAngle - in Radians *)

function GetX: integer;
function GetY: integer;
procedure MoveTo(x, y: integer);
procedure MoveRel(xshift, yshift: integer);
procedure LineRel(x, y: integer);
procedure LineTo(x, y: integer);

function GetBkColor: TColorRef;
procedure SetSolidBrush;
procedure SetNullBrush;
procedure SetHatchedBrush(Hatch: DWord);
(*
HS_BDIAGONAL	A 45-degree upward, left-to-right hatch
HS_CROSS	Horizontal and vertical cross-hatch
HS_DIAGCROSS	45-degree crosshatch
HS_FDIAGONAL	A 45-degree downward, left-to-right hatch
HS_HORIZONTAL	Horizontal hatch
HS_VERTICAL	Vertical hatch
*)
procedure SetBrushColor(Color: TColorRef);

procedure FloodFillSurface(X, Y: Integer; Color: TColorRef);
procedure FloodFillBorder(X, Y: Integer; Color: TColorRef);

procedure SetTextStyle(FontName: String; Height: Integer; Weight: Integer; Italic: Boolean = False; Underline: Boolean = False; StrikeOut: Boolean = False; Escapement: Integer = 0);
(* Escapement - in tenth of degrees, i.e. 450 = 45 degrees*)
(*
FW_DONTCARE	0
FW_THIN	100
FW_EXTRALIGHT	200
FW_ULTRALIGHT	200
FW_LIGHT	300
FW_NORMAL	400
FW_REGULAR	400
FW_MEDIUM	500
FW_SEMIBOLD	600
FW_DEMIBOLD	600
FW_BOLD	700
FW_EXTRABOLD	800
FW_ULTRABOLD	800
FW_HEAVY	900
FW_BLACK	900
*)
procedure SetTextColor(Color: TColorRef);
procedure SetTextAlign(Flags: UINT);
(*
TA_BASELINE	The reference point will be on the base line of the text.
TA_BOTTOM	The reference point will be on the bottom edge of the bounding rectangle.
TA_TOP	The reference point will be on the top edge of the bounding rectangle.
TA_CENTER	The reference point will be aligned horizontally with the center of the bounding rectangle.
TA_LEFT	The reference point will be on the left edge of the bounding rectangle.
TA_RIGHT	The reference point will be on the right edge of the bounding rectangle.
*)
procedure OutTextXY(x, y: integer; text: string);

Function RePaint: Bool;

(*   WinAPI- *)

Function GetGraphDC: HDC;

(*   *)

Function ReadKey: TKeyEventRecord;

implementation

(*   *)

const
  GraphWindowClassName = 'TGraphWindow';
  GraphWindowWindowName = 'Graph';

  WindowedWindowStyle = WS_POPUP or WS_THICKFRAME or WS_CAPTION or WS_SYSMENU or WS_MINIMIZEBOX or WS_MAXIMIZEBOX;
  NoBorderWindowStyle = WS_POPUP;

var
  GraphThread: THandle;           //      
  GraphWindow: HWND;              //  
  GraphDC: HDC;                   // DC,     
  GraphWindowClass: TAtom;        //   
  GraphInitializedEvent: THandle; //    GraphWindow  ..?

(*   *)

Var
  ScreenWidth, ScreenHeight: DWord;
  DesktopRect: TRect;

  KeyEvent: TKeyEventRecord;
  KeyState: TKeyboardState;
  KeyEventReady: THandle;

(*   *)


Function ReadKey: TKeyEventRecord;
var
  Handles: Array [1..2] of THandle;
begin
  Handles[1] := KeyEventReady;
  Handles[2] := GraphThread;
//  WaitForSingleObject(KeyEventReady, INFINITE);
  WaitForMultipleObjects(2, @Handles[1], False, INFINITE);
  ResetEvent(KeyEventReady);
End;


Function SetGraphResolution(W, H: DWord): Boolean;
var
  Rect: TRect;
begin
  //    
  Rect.Left := (DesktopRect.Right - DesktopRect.Left - Integer(W)) div 2;
  if (Rect.Left < 0) then
    Rect.Left := DesktopRect.Left;

  Rect.Top := (DesktopRect.Bottom - DesktopRect.Top - Integer(H)) div 2;
  if (Rect.Top < 0) then
    Rect.Top := DesktopRect.Top;

  Rect.Right := Rect.Left + Integer(W);
  Rect.Bottom := Rect.Top + Integer(H);

  AdjustWindowRectEx(Rect, GetWindowLong(GraphWindow, GWL_STYLE), False, GetWindowLong(GraphWindow, GWL_EXSTYLE));

  Result := SetWindowPos(GraphWindow, 0, Rect.Left, Rect.Top, Rect.Right - Rect.Left, Rect.Bottom - Rect.Top, SWP_NOZORDER);
End;

Function SetGraphToDesktop: Boolean;
begin
  Result := SetGraphResolution(DesktopRect.Right - DesktopRect.Left, DesktopRect.Bottom - DesktopRect.Top);
End;

Function SetGraphToScreen: Boolean;
var
  ScreenRect, SaveDeskRect: TRect;
begin
  SaveDeskRect := DesktopRect;

  ScreenRect.Left := 0;
  ScreenRect.Top := 0;
  ScreenRect.Right := ScreenWidth;
  ScreenRect.Bottom := ScreenHeight;

  DesktopRect := ScreenRect;

  Result := SetGraphResolution(ScreenWidth, ScreenHeight);

  DesktopRect := SaveDeskRect;
End;


Function SetWindowedMode(Windowed: Boolean): Boolean;
var w, h: DWord;
begin
  w := WorkWidth;
  h := WorkHeight;

  Result := Windowed;
  if Windowed then
    SetWindowLong(GraphWindow, GWL_STYLE, Integer(WindowedWindowStyle)) else
    SetWindowLong(GraphWindow, GWL_STYLE, Integer(NoBorderWindowStyle));

  SetGraphResolution(w, h);  
End;

Function SetOnTopMode(OnTop: Boolean): Boolean;
begin
  if OnTop then
    Result := SetWindowPos(GraphWindow, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOMOVE or SWP_NOSIZE) else
    Result := SetWindowPos(GraphWindow, HWND_NOTOPMOST, 0, 0, 0, 0, SWP_NOMOVE or SWP_NOSIZE);
End;


function SetFullScreenMode(W, H: UInt; C: Byte): Bool;
var
  DeviceMode: TDevMode;
begin
  with DeviceMode do
  begin
    dmSize := SizeOf(DeviceMode);
    dmBitsPerPel := C;
    dmPelsWidth := W;
    dmPelsHeight := H;
    dmFields := DM_PELSWIDTH or DM_PELSHEIGHT;
    if (C <> 0) then
      dmFields := dmFields or DM_BITSPERPEL;
  enD;

  Result :=
    (ChangeDisplaySettings(DeviceMode, CDS_TEST or CDS_FULLSCREEN) = DISP_CHANGE_SUCCESSFUL) and
    (ChangeDisplaySettings(DeviceMode, CDS_FULLSCREEN) = DISP_CHANGE_SUCCESSFUL);
end;

function ClearFullScreenMode: Bool;
begin
  Result :=
    (ChangeDisplaySettings(pDevMode(nil)^, CDS_FULLSCREEN) = DISP_CHANGE_SUCCESSFUL);
enD;

Function SetFullScreenMode(FullScreen: Boolean): Boolean;
begin
  if (FullScreen) then
    Result := SetFullScreenMode(WorkWidth, WorkHeight) and (SetGraphToScreen) else
    Result := ClearFullScreenMode;
End;


Function ShowGraphWindow: Boolean;
begin
  ShowWindow(GraphWindow, SW_SHOW);
  Result := SetForegroundWindow(GraphWindow);
End;

Function HideGraphWindow: Boolean;
begin
  SetFullScreenMode(False);
  ShowWindow(GraphWindow, SW_HIDE);
  Result := True;
End;


Function WaitGraph: Boolean;
begin
  Result := False;
  
  if (GraphWindow <> 0) and
     (IsWindowVisible(GraphWindow))
  then
    begin
      RePaint;
      Result := (WaitForSingleObject(GraphThread, INFINITE) <> WAIT_FAILED);
    End;
End;

(*   *)

Function WorkWidth: DWord;
var
  Rect: TRect;
begin
  GetClientRect(GraphWindow, Rect);

  Result := Rect.Right - Rect.Left;
End;

Function WorkHeight: DWord;
var
  Rect: TRect;
begin
  GetClientRect(GraphWindow, Rect);

  Result := Rect.Bottom - Rect.Top;
End;

function GetPixel(X, Y: Integer): TColorRef;
begin
  Result := Windows.GetPixel(GraphDC, X, Y);
End;

function SetPixel(X, Y: Integer; Color: TColorRef): TColorRef; // ~ PutPixel
begin
  Result := Windows.SetPixel(GraphDC, X, Y, Color);
End;

procedure SetLineWidth(Width: DWord);
var
  Pen: HPen;
  LogPen: TLogPen;
begin
  Pen := GetCurrentObject(GraphDC, OBJ_PEN);
  GetObject(Pen, SizeOf(LogPen), @LogPen);

  LogPen.lopnWidth.X := Width;

  Pen := SelectObject(GraphDC, CreatePenIndirect(LogPen));
  DeleteObject(Pen);
End;

procedure SetLineStyle(Style: DWord);
var
  Pen: HPen;
  LogPen: TLogPen;
begin
  Pen := GetCurrentObject(GraphDC, OBJ_PEN);
  GetObject(Pen, SizeOf(LogPen), @LogPen);

  LogPen.lopnStyle := Style;
  LogPen.lopnWidth.X := 1;

  Pen := SelectObject(GraphDC, CreatePenIndirect(LogPen));
  DeleteObject(Pen);
End;

procedure SetLineColor(Color: TColorRef);
var
  Pen: HPen;
  LogPen: TLogPen;
begin
  Pen := GetCurrentObject(GraphDC, OBJ_PEN);
  GetObject(Pen, SizeOf(LogPen), @LogPen);

  LogPen.lopnColor := Color;

  Pen := SelectObject(GraphDC, CreatePenIndirect(LogPen));
  DeleteObject(Pen);
End;

function GetLineColor: TColorRef;
var
  Pen: HPen;
  LogPen: TLogPen;
begin
  Pen := GetCurrentObject(GraphDC, OBJ_PEN);
  GetObject(Pen, SizeOf(LogPen), @LogPen);

  Result := LogPen.lopnColor;
End;

procedure Line(x1, y1, x2, y2: integer);
var
  LastPos: TPoint;
begin
  Windows.MoveToEx(GraphDC, x1, y1, @LastPos);
  Windows.LineTo(GraphDC, x2, y2);
  Windows.MoveToEx(GraphDC, LastPos.X, LastPos.Y, nil);
End;

procedure Polyline(var Points; Count: Integer);
begin
  Windows.Polyline(GraphDC, Points, Count);
End;

procedure PolyBezier(const Points; Count: DWORD);
begin
  Windows.PolyBezier(GraphDC, Points, Count);
End;

procedure Circle(x, y, r: integer);
begin
  Windows.Ellipse(GraphDC, x - r, y - r, x + r, y + r);
End;

procedure Polygon(var Points; Count: Integer);
begin
  Windows.Polygon(GraphDC, Points, Count);
End;

procedure Rectangle(x1, y1, x2, y2: integer);
begin
  Windows.Rectangle(GraphDC, x1, y1, x2, y2);
End;

procedure RoundRect(x1, y1, x2, y2, rx, ry: integer);
begin
  Windows.RoundRect(GraphDC, x1, y1, x2, y2, rx, ry);
End;

procedure Ellipse(x, y: integer; xRadius, yRadius: integer);
begin
  Windows.Ellipse(GraphDC, x - xRadius, y - yRadius, x + xRadius, y + yRadius);
End;

procedure Arc(X, Y: Integer; xRadius, yRadius: Integer; StartAngle, EndAngle: Single);
begin
  Windows.Arc(GraphDC,
    X - xRadius, Y - yRadius, X + xRadius, Y + yRadius,
    Round(X + xRadius*Cos(StartAngle)), Round(Y - yRadius*Sin(StartAngle)), Round(X + xRadius*Cos(EndAngle)), Round(Y - yRadius*Sin(EndAngle)));
End;

procedure Pie(X, Y: Integer; xRadius, yRadius: Integer; StartAngle, EndAngle: Single);
begin
  Windows.Pie(GraphDC,
    X - xRadius, Y - yRadius, X + xRadius, Y + yRadius,
    Round(X + xRadius*Cos(StartAngle)), Round(Y - yRadius*Sin(StartAngle)), Round(X + xRadius*Cos(EndAngle)), Round(Y - yRadius*Sin(EndAngle)));
End;

procedure Chord(X, Y: Integer; xRadius, yRadius: Integer; StartAngle, EndAngle: Single);
begin
  Windows.Chord(GraphDC,
    X - xRadius, Y - yRadius, X + xRadius, Y + yRadius,
    Round(X + xRadius*Cos(StartAngle)), Round(Y - yRadius*Sin(StartAngle)), Round(X + xRadius*Cos(EndAngle)), Round(Y - yRadius*Sin(EndAngle)));
End;

function GetX: integer;
var
  LastPos: TPoint;
begin
  MoveToEx(GraphDC, 0, 0, @LastPos);
  MoveToEx(GraphDC, LastPos.X, LastPos.Y, nil);
  GetX := LastPos.X;
End;

function GetY: integer;
var
  LastPos: TPoint;
begin
  MoveToEx(GraphDC, 0, 0, @LastPos);
  MoveToEx(GraphDC, LastPos.X, LastPos.Y, nil);
  GetY := LastPos.Y;
End;

procedure MoveTo(x, y: integer);
begin
  MoveToEx(GraphDC, x, y, nil);
End;

procedure MoveRel(xshift, yshift: integer);
var
  LastPos: TPoint;
begin
  MoveToEx(GraphDC, 0, 0, @LastPos);
  MoveToEx(GraphDC, LastPos.X + xshift, LastPos.Y + yshift, nil);
End;

procedure LineTo(x, y: integer);
begin
  Windows.LineTo(GraphDC, x, y);
End;

procedure LineRel(x, y: integer);
var
  LastPos: TPoint;
begin
  MoveToEx(GraphDC, 0, 0, @LastPos);
  MoveToEx(GraphDC, LastPos.X, LastPos.Y, nil);
  Windows.LineTo(GraphDC, LastPos.X + x, LastPos.Y + y);
End;

function GetBkColor: TColorRef;
begin
  Result := Windows.GetBkColor(GraphDC);
End;

procedure SetSolidBrush;
var
  Brush: HBRUSH;
  LogBrush: TLogBrush;
begin
  Brush := GetCurrentObject(GraphDC, OBJ_BRUSH);
  GetObject(Brush, SizeOf(LogBrush), @LogBrush);

  LogBrush.lbStyle := BS_SOLID;
  SetBkMode(GraphDC, OPAQUE);

  Brush := SelectObject(GraphDC, CreateBrushIndirect(LogBrush));
  DeleteObject(Brush);
End;

procedure SetNullBrush;
var
  Brush: HBRUSH;
  LogBrush: TLogBrush;
begin
  Brush := GetCurrentObject(GraphDC, OBJ_BRUSH);
  GetObject(Brush, SizeOf(LogBrush), @LogBrush);

  LogBrush.lbStyle := BS_NULL;
  SetBkMode(GraphDC, TRANSPARENT);

  Brush := SelectObject(GraphDC, CreateBrushIndirect(LogBrush));
  DeleteObject(Brush);
End;

procedure SetHatchedBrush(Hatch: DWord);
var
  Brush: HBRUSH;
  LogBrush: TLogBrush;
begin
  Brush := GetCurrentObject(GraphDC, OBJ_BRUSH);
  GetObject(Brush, SizeOf(LogBrush), @LogBrush);

  LogBrush.lbStyle := BS_HATCHED;
  LogBrush.lbHatch := Hatch;

  Brush := SelectObject(GraphDC, CreateBrushIndirect(LogBrush));
  DeleteObject(Brush);
End;

procedure SetBrushColor(Color: TColorRef);
var
  Brush: HBRUSH;
  LogBrush: TLogBrush;
begin
  Brush := GetCurrentObject(GraphDC, OBJ_BRUSH);
  GetObject(Brush, SizeOf(LogBrush), @LogBrush);

  LogBrush.lbColor := Color;
  SetBkColor(GraphDC, Color);

  Brush := SelectObject(GraphDC, CreateBrushIndirect(LogBrush));
  DeleteObject(Brush);
End;


procedure FloodFillSurface(X, Y: Integer; Color: TColorRef);
begin
  ExtFloodFill(GraphDC, X, Y, Color, Windows.FLOODFILLSURFACE);
End;

procedure FloodFillBorder(X, Y: Integer; Color: TColorRef);
begin
  ExtFloodFill(GraphDC, X, Y, Color, Windows.FLOODFILLBORDER);
End;

procedure SetTextStyle(FontName: String; Height: Integer; Weight: Integer; Italic: Boolean = False; Underline: Boolean = False; StrikeOut: Boolean = False; Escapement: Integer = 0);
var
  LogFont: TLogFont;
begin
  LogFont.lfHeight := Height;
  LogFont.lfWidth := 0;
  LogFont.lfEscapement := Escapement;
  LogFont.lfOrientation := 0;
  LogFont.lfWeight := Weight;
  LogFont.lfItalic := Byte(Italic);
  LogFont.lfUnderline := Byte(Underline);
  LogFont.lfStrikeOut := Byte(StrikeOut);
  LogFont.lfCharSet := DEFAULT_CHARSET;
  LogFont.lfOutPrecision := OUT_DEFAULT_PRECIS;
  LogFont.lfClipPrecision := CLIP_DEFAULT_PRECIS;
  LogFont.lfQuality := PROOF_QUALITY;
  LogFont.lfPitchAndFamily := FF_DONTCARE or DEFAULT_PITCH; // 0 =))
  StrCopy(LogFont.lfFaceName, PChar(Copy(FontName, 1, 32)));
//  LogFont.lfFaceName := FontName;

  SelectObject(GraphDC, CreateFontIndirect(LogFont));
End;

procedure SetTextColor(Color: TColorRef);
begin
  Windows.SetTextColor(GraphDC, Color);
End;

procedure SetTextAlign(Flags: UINT);
begin
  Windows.SetTextAlign(GraphDC, Flags);
End;

procedure OutTextXY(x, y: integer; text: string);
begin
(*  DrawTextEx *)
  TextOut(GraphDC, X, Y, PChar(Text), Length(Text));
End;

Function RePaint: Bool;
begin
  Result := InvalidateRect(GraphWindow, nil, False);
End;

Function GetGraphDC: HDC;
begin
  Result := GraphDC;
End;

(*   *)

// Main Window Procedure for Graph Window
function WindowProc(hWnd: HWND; Msg: UINT; wParam: WPARAM; lParam: LPARAM): LRESULT; stdcall;
var
  DC, OldDC: HDC;
  GraphBM, OldBM: HBitMap;
  BMInfo: TBitmap;
  Paint: TPaintStruct;
  ConIn: THandle;
  Rec: TInputRecord;
  Written: DWord;
  
begin

  Case Msg of
    WM_PAINT:
      begin
        BeginPaint(hWnd, Paint);

        BitBlt(Paint.hdc, 0, 0, WorkWidth, WorkHeight, GraphDC, 0, 0, SRCCOPY);

        EndPaint(hWnd, Paint);
        Result := 0;
      End;

    WM_ERASEBKGND:
      begin
        Result := 0;
      End;

    WM_SIZE:
      begin
        // Create DC
        DC := GetDC(hWnd);

        OldDC := GraphDC;
        GraphDC := CreateCompatibleDC(DC);
        GraphBM := CreateCompatibleBitMap(DC, LOWORD(lParam), HIWORD(lParam) (* WorkWidth, WorkHeight *));
        SelectObject(GraphDC, GraphBM);

        // Copy Old Content
        if (OldDC <> 0) then
          begin
            OldBM := GetCurrentObject(OldDC, OBJ_BITMAP);
            GetObject(OldBM, SizeOf(BMInfo), @BMInfo);

            BitBlt(GraphDC, 0, 0, BMInfo.bmWidth, BMInfo.bmHeight, OldDC, 0, 0, SRCCOPY);

            DeleteObject(OldBM);
            DeleteDC(OldDC);
          End;

        ReleaseDC(hWnd, DC);

        Result := 0;
      End;

    WM_CLOSE:
      begin
        DestroyWindow(hWnd);
        Result := 0;
      End;

    WM_KEYDOWN, WM_KEYUP:
      begin
(*
        nVirtKey = (int) wParam;    // virtual-key code
        lKeyData = lParam;          // key data
*)

        KeyEvent.bKeyDown := (Msg = WM_KEYDOWN);
        KeyEvent.wRepeatCount := lParam and $FFFF; // 0-15 Specifies the repeat count.
        KeyEvent.wVirtualKeyCode := wParam;
        KeyEvent.wVirtualScanCode := (lParam shr 16) and $FF; // 16-23 Specifies the scan code.

        GetKeyboardState(KeyState);

//        KeyEvent.AsciiChar := ...;
        ToAscii(KeyEvent.wVirtualKeyCode, KeyEvent.wVirtualScanCode, KeyState, @KeyEvent.AsciiChar, 0);

//        ToUnicode(KeyEvent.wVirtualKeyCode, KeyEvent.wVirtualScanCode, KeyState, , 0);
//        KeyEvent.UnicodeChar := ...;
//        KeyEvent.dwControlKeyState := ??? !!! ;

        ConIn := GetStdHandle(STD_INPUT_HANDLE);

        if (ConIn <> INVALID_HANDLE_VALUE) and (ConIn <> 0) then
          begin
            Rec.EventType := KEY_EVENT;
            Rec.Event.KeyEvent := KeyEvent;

            WriteConsoleInput(ConIn, Rec, SizeOf(Rec), Written);
          end else begin
            SetEvent(KeyEventReady);
          enD;
      End;
(*
    WM_KEYDOWN:
      begin
        if (wParam = 27) then // ESC Pressed
          DestroyWindow(hWnd);

        if (wParam = 112) then // F1 Pressed
          begin
            MessageBox(hWnd, PChar('Resolution is: (' + IntToStr(WorkWidth) + ', ' + IntToStr(WorkHeight) + ')'), 'Info', MB_OK);
          End;

        Result := 0;
      End;
*)

    WM_DESTROY:
      begin
        // Clearing
        OldBM := GetCurrentObject(GraphDC, OBJ_BITMAP);
        DeleteObject(OldBM);

        DeleteDC(GraphDC);
        GraphDC := 0;

        PostQuitMessage(0);
        Result := 0;
      End;

    else
      Result := DefWindowProc(hWnd, Msg, wParam, lParam);
    End;
End;

// Main Procedure for graph window Thread
Function GraphThreadFunc(Param: DWORD): DWORD; stdcall;
var
  Msg: TMsg;
  WndClass: TWndClassEx;
begin
  // Register Class
  with WndClass do
    begin
      cbSize := SizeOf(WndClass);
      style := 0;
      lpfnWndProc := @WindowProc;
      cbClsExtra := 0;
      cbWndExtra := 0;
      hInstance := SysInit.HInstance; //GetModuleHandle(0);
      hIcon := LoadIcon(0, IDI_APPLICATION);
      hCursor := LoadCursor(0, IDC_ARROW);
      hbrBackground := GetStockObject(BLACK_BRUSH);
      lpszMenuName := nil;
      lpszClassName := GraphWindowClassName;
      hIconSm := 0;
    End;

  GraphWindowClass := RegisterClassEx(WndClass);

  if (GraphWindowClass <> 0) then
    begin
      GraphDC := 0;

      // Create Window
      GraphWindow :=
        CreateWindowEx
          ( WS_EX_APPWINDOW,
            GraphWindowClassName,
            GraphWindowWindowName,
            NoBorderWindowStyle,
            0, 0, 0, 0,
            0, 0, hInstance, nil (* Parameter *)
          );
    End;

  if (GraphWindow <> 0) then
    SetGraphResolution(DefaultWidth, DefaultHeight);

  // Here Graph is initialized.
  SetEvent(GraphInitializedEvent);

  // Initialize Result
  Result := 0;

  // Try to process message Loop
  if (GraphWindow <> 0) then
    begin
      // Start the message loop.
      while (GetMessage(Msg, 0, 0, 0)) do
        begin
          TranslateMessage(Msg);
          DispatchMessage(Msg);
        End;

      // Return the exit code to Windows.
      Result := Msg.wParam;
    End;
End;

// Initialization of graph Thread 
Function InitGraph: Boolean;
var
  ThreadID: DWORD;
begin
  // Create Event
  GraphInitializedEvent := CreateEvent(nil, True, False, nil);
  KeyEventReady := CreateEvent(nil, True, False, nil);

  // Some Initialization
  ScreenWidth := GetSystemMetrics(SM_CXSCREEN);
  ScreenHeight := GetSystemMetrics(SM_CYSCREEN);

  SystemParametersInfo(SPI_GETWORKAREA, 0, @DesktopRect, 0);

  // Create Thread
  GraphThread := CreateThread(nil, 0, @GraphThreadFunc, nil (* parameter *), 0, ThreadID);

  // Wait for Ready...
  Result := (WaitForSingleObject(GraphInitializedEvent, INFINITE) <> WAIT_FAILED);
End;

Function CloseGraph: Bool;
begin
  if (GraphWindow <> 0) then
    begin
      SendMessage(GraphWindow, WM_DESTROY, 0, 0);
      GraphWindow := 0;
    End;
End;

initialization
  InitGraph;
finalization
  CloseGraph;
end.
